/*
* This program and the accompanying materials are made available under the terms of the
* Eclipse Public License v2.0 which accompanies this distribution, and is available at
* https://www.eclipse.org/legal/epl-v20.html
*
* SPDX-License-Identifier: EPL-2.0
*
* Copyright Contributors to the Zowe Project.
*
*/

jest.mock("../../../src/Imperative");
import { IProfile, IProfileLoaded } from "../../../../profiles";
import { CommandProfiles, IHandlerParameters, ICommandDefinition } from "../../../../cmd";
import { Imperative } from "../../../src/Imperative";

// "Mocked" profiles
const FAKE_PROFS: IProfileLoaded[] = [
    {
        message: "loaded",
        type: "fake",
        failNotFound: true,
        name: "fake1",
        profile: {
            name: "fake1",
            type: "fake",
            data: "some data",
            info: "some info",
            nested: {
                data: "nested data"
            }
        }
    },
    {
        message: "loaded",
        type: "fake",
        failNotFound: true,
        name: "fake",
        profile: {
            name: "fake2",
            type: "fake",
            data: "some data",
            info: "some info",
            nested: {
                data: "nested data"
            }
        }
    }
];

// "Mocked" version of the imperative API - done here rather than a manual mock
const MockedImperativeAPI = {
    profileManager: (args) => {
        return {
            getDefaultProfileName: jest.fn(() => {
                return "fake1";
            }),
            loadAll: jest.fn((loadArgs) => {
                return JSON.parse(JSON.stringify(FAKE_PROFS));
            })
        };
    },
};

// "Mocked" version to thrown an error
const MockedImperativeAPIError = {
    profileManager: (args) => {
        return {
            getDefaultProfileName: jest.fn(() => {
                return "fake1";
            }),
            loadAll: jest.fn((loadArgs) => {
                throw new Error("ERROR!");
            })
        };
    },
};

// "Mocked" version of the handler parameters
const HANDLER_PARAMETERS: any = {
    arguments: {
        $0: "bright",
        _: ["zos-jobs", "submit", "data-set"],
    },
    response: {
        data: {
            setMessage: jest.fn((setMsgArgs) => {
                // Nothing
            }),
            setObj: jest.fn((setObjArgs) => {
                // Nothing
            })
        },
        console: {
            log: jest.fn((logs) => {
                // Nothing
            }),
            error: jest.fn((errors) => {
                // Nothing
            }),
            errorHeader: jest.fn(() => undefined)
        },
        progress: {
            startBar: jest.fn((parms) => undefined),
            endBar: jest.fn(() => undefined)
        },
        format: {
            output: jest.fn((parms) => {
                // Nothing
            })
        }
    },
    definition: {
        customize: {
            profileTypeIdentifier: "fake"
        }
    },
    fullDefinition: undefined,
    profiles: undefined
};

describe("list profiles handler", () => {

    // Reset mocks for counters, etc.
    afterEach(() => {
        jest.resetAllMocks();
    });

    describe("error handling", () => {
        it("should not transform errors from the profile manager", async () => {
            // "Mock" with the object here
            Object.defineProperty(Imperative, "api", { value: MockedImperativeAPIError, configurable: true });
            const handlerReq = require("../../../src/profiles/handlers/ListProfilesHandler");
            const handler = new handlerReq.default();
            const parms = Object.assign({}, ...[HANDLER_PARAMETERS]);
            let error;
            try {
                await handler.process(parms);
            } catch (err) {
                error = err;
            }
            expect(error).toBeDefined();
            expect(error.message).toMatchSnapshot();
        });
    });

    describe("response", () => {
        it("should load all profiles and display just the names", async () => {
            // "Mock" with the object here
            Object.defineProperty(Imperative, "api", { value: MockedImperativeAPI, configurable: true });
            const handlerReq = require("../../../src/profiles/handlers/ListProfilesHandler");
            const handler = new handlerReq.default();
            const parms = Object.assign({}, ...[HANDLER_PARAMETERS]);
            parms.response.format.output = jest.fn((args) => {
                expect(args).toMatchSnapshot();
            });
            parms.response.data.setObj = jest.fn((args) => {
                expect(args).toMatchSnapshot();
            });
            parms.response.data.setMessage = jest.fn((args) => {
                expect(args).toMatchSnapshot();
            });
            await handler.process(parms);
            expect(parms.response.data.setMessage).toHaveBeenCalledTimes(1);
            expect(parms.response.data.setObj).toHaveBeenCalledTimes(1);
        });

        it("should load all profiles and display all contents", async () => {
            // "Mock" with the object here
            Object.defineProperty(Imperative, "api", { value: MockedImperativeAPI, configurable: true });
            const handlerReq = require("../../../src/profiles/handlers/ListProfilesHandler");
            const handler = new handlerReq.default();
            const parms = Object.assign({}, ...[HANDLER_PARAMETERS]);
            parms.arguments.showContents = true;
            parms.response.format.output = jest.fn((args) => {
                expect(args).toMatchSnapshot();
            });
            parms.response.data.setObj = jest.fn((args) => {
                expect(args).toMatchSnapshot();
            });
            parms.response.data.setMessage = jest.fn((args) => {
                expect(args).toMatchSnapshot();
            });
            await handler.process(parms);
            expect(parms.response.data.setMessage).toHaveBeenCalledTimes(1);
            expect(parms.response.data.setObj).toHaveBeenCalledTimes(1);
        });
    });
});
